@@ -4,10 +4,78 @@ import subprocess
 import yaml
 import sys
 
+def error_text(s):
+    return f'**{s}**'
+
+class SVInstParsingError(Exception):
+    """Exception raised for errors while parsing.
+
+    Attributes:
+        error_set: Set of file indices that have parsing errors (starts at zero)
+        file_list: List of files that were passed to svinst
+        svinst_stderr: Output of the svinst tool to standard error (stderr)
+        svinst_retcode: Return code of svinst
+    """
+
+    def __init__(self, error_set=None, file_list=None, svinst_stderr=None,
+                 svinst_retcode=None, nl='\n'):
+        # save information
+        self.error_set = error_set
+        self.file_list = file_list
+        self.svinst_stderr = svinst_stderr
+        self.svinst_retcode = svinst_retcode
+
+        # render message
+        message = []
+        if svinst_stderr is not None:
+            message += [f'{svinst_stderr}']
+        if (error_set is not None) and (file_list is not None):
+            # render a list of indices
+            # make a list of the file indices with errors, starting
+            # at "1" to make it more human-readable
+            indices = list(error_set)
+            indices = sorted(indices)
+            indices = [str(x+1) for x in indices]
+            indices = ', '.join(indices)
+
+            # mark the files with errors
+            files_str = []
+            for k, file in enumerate(file_list):
+                name = Path(file).name
+                if k in error_set:
+                    name = error_text(name)
+                files_str.append(name)
+            marked_list = '[' + ', '.join(files_str) + ']'
+
+            # create the error message
+            message += [f'Error in file(s) {indices}: {marked_list}']
+        if svinst_retcode is not None:
+            message += [f'svinst returned code {svinst_retcode}.']
+
+        super().__init__(nl.join(message))
+
 def is_single_file(files):
     return isinstance(files, (str, Path))
 
-def call_svinst(files, includes=None, defines=None, ignore_include=False, full_tree=False):
+def find_error_indices(files, stdout):
+    # try to parse the YAML output
+    d = yaml.safe_load(stdout)
+
+    # figure out which files had errors (i.e., which are missing)
+    error_set = set()
+    idx = 0
+    for k in range(len(files)):
+        if (idx < len(d['files']) and
+            (str(d['files'][idx]['file_name']) == str(files[k]))):
+            idx += 1
+        else:
+            error_set.add(k)
+
+    # return the indices
+    return error_set
+
+def call_svinst(files, includes=None, defines=None, ignore_include=False, full_tree=False,
+                separate=False, show_macro_defs=False, explain_error=True):
     # set defaults
     if includes is None:
         includes = []
@@ -32,6 +100,10 @@ def call_svinst(files, includes=None, defines=None, ignore_include=False, full_t
         args += ['--ignore-include']
     if full_tree:
         args += ['--full-tree']
+    if separate:
+        args += ['--separate']
+    if show_macro_defs:
+        args += ['--show-macro-defs']
 
     # convert arguments to strings
     args = [str(elem) for elem in args]
@@ -39,8 +111,17 @@ def call_svinst(files, includes=None, defines=None, ignore_include=False, full_t
     # call command
     result = subprocess.run(args, capture_output=True, encoding='utf-8')
     if result.returncode != 0:
-        print(f'{result.stderr}', file=sys.stderr)
-        raise Exception(f'svinst returned code {result.returncode}.')
+        # find indices of files with errors, if desired
+        error_set = None
+        if explain_error and (len(files) > 1):
+            try:
+                error_set = find_error_indices(files=files, stdout=result.stdout)
+            except:
+                pass
+
+        # raise an exception
+        raise SVInstParsingError(error_set=error_set,file_list=files,
+                                 svinst_stderr=result.stderr, svinst_retcode=result.returncode)
 
     # parse output as YAML
     return yaml.safe_load(result.stdout)
@@ -69,6 +150,17 @@ class Def:
                 (len(self.insts) == len(other.insts)) and
                 all(self_inst == other_inst for self_inst, other_inst in zip(self.insts, other.insts)))
 
+class MacroDef:
+    def __init__(self, text):
+        self.text = text
+
+    def __str__(self):
+        return f"{self.__class__.__name__}('{self.text}')"
+
+    def __eq__(self, other):
+        return (isinstance(other, self.__class__) and
+                (self.text == other.text))
+
 class ModDef(Def):
     pass
 
@@ -106,6 +198,9 @@ class PkgInst(Inst):
 def process_defs(result):
     retval = []
 
+    if result is None:
+        result = []
+
     for entry in result:
         if 'pkg_name' in entry:
             def_ = PkgDef(name=entry['pkg_name'])
@@ -129,13 +224,34 @@ def process_defs(result):
 
     return retval
 
-def get_defs(files, includes=None, defines=None, ignore_include=False):
+def process_macro_defs(result):
+    retval = []
+
+    if result is None:
+        result = []
+
+    for entry in result:
+        retval.append(MacroDef(text=entry))
+
+    return retval
+
+def get_defs(files, includes=None, defines=None, ignore_include=False,
+             separate=False, show_macro_defs=False, explain_error=True):
     single = is_single_file(files)
 
     out = call_svinst(files=files, includes=includes, defines=defines,
-                      ignore_include=ignore_include, full_tree=False)
+                      ignore_include=ignore_include, separate=separate,
+                      show_macro_defs=show_macro_defs, explain_error=explain_error,
+                      full_tree=False)
 
-    retval = [process_defs(elem['defs']) for elem in out['files']]
+    retval = []
+    for elem in out['files']:
+        retval.append([])
+        retval[-1].extend(process_defs(elem['defs']))
+        if show_macro_defs:
+            macro_defs = process_macro_defs(elem['macro_defs'])
+            macro_defs = sorted(macro_defs, key=lambda x: x.text)
+            retval[-1].extend(macro_defs)
 
     if single:
         retval = retval[0]
@@ -209,11 +325,14 @@ def process_syntax_tree(result):
             )
     return retval
 
-def get_syntax_tree(files, includes=None, defines=None, ignore_include=False):
+def get_syntax_tree(files, includes=None, defines=None, ignore_include=False,
+                    separate=False, show_macro_defs=False, explain_error=True):
     single = is_single_file(files)
 
     out = call_svinst(files=files, includes=includes, defines=defines,
-                      ignore_include=ignore_include, full_tree=True)
+                      ignore_include=ignore_include, separate=separate,
+                      show_macro_defs=show_macro_defs, explain_error=explain_error,
+                      full_tree=True)
 
     retval = [process_syntax_tree(elem['syntax_tree']) for elem in out['files']]
 
